#include "precompile.h"

#include <mutex>

#include <google/protobuf/message.h>
#include <a8/websocketsession.h>
#include <a8/tcplistener.h>

#include <f8/netmsghandler.h>
#include <f8/udplog.h>
#include <f8/msgqueue.h>

#include "app.h"
#include "GCListener.h"
#include "jsondatamgr.h"
#include "ss_proto.pb.h"
#include "handlermgr.h"

class GCClientSession: public a8::WebSocketSession
{
public:

    virtual void DecodeUserPacket(char* buf, int& offset, unsigned int buflen) override
    {
        //packagelen + msgid + magiccode + msgbody
        //2 + 2 + 4+ xx + \0 + xx
        bool warning = false;
        while (buflen - offset >= sizeof(f8::PackHead)) {
            f8::PackHead* p = (f8::PackHead*)&buf[offset];
            if (p->magic_code == f8::MAGIC_CODE) {
                if (buflen - offset < sizeof(f8::PackHead) + p->packlen) {
                    break;
                }
                App::Instance()->AddSocketMsg(SF_Client,
                                              socket_handle,
                                              saddr,
                                              p->msgid,
                                              p->seqid,
                                              &buf[offset + sizeof(f8::PackHead)],
                                              p->packlen,
                                              ST_Tcp);
                offset += sizeof(f8::PackHead) + p->packlen;
            } else {
                warning = true;
                offset++;
                continue;
            }
        }

        if (warning) {
            f8::UdpLog::Instance()->Warning("收到client非法数据包", {});
        }
    }

    virtual void OnRawHttpGet(const std::string& url, const std::string& querystr,
                              std::string& response) override
    {
        f8::MsgQueue::Instance()->PostMsg
            (
             IM_ExecGM,
             a8::Args
             (
              {
               (int)socket_handle,
               (std::string)(url + ""),
               (std::string)(querystr + ""),
               (unsigned long)saddr
              }
              )
             );
    }

    virtual bool HandleRedirect(const std::string& url, const std::string& querystr,
                                std::string& location) override
    {
#ifdef DEBUG
        f8::MsgQueue::Instance()->PostMsg
            (
             IM_HandleRedirect,
             a8::Args
             (
              {
                  (int)socket_handle,
                      (std::string)(url + ""),
                      (std::string)(querystr + ""),
                      (unsigned long)saddr
              }
              )
             );
#endif
        return false;
    }

    virtual void OnDisConnect() override
    {
        f8::MsgQueue::Instance()->PostMsg
            (
             IM_ClientSocketDisconnect,
             a8::Args
             (
              {
               (int)socket_handle
              }
              )
             );
    }

};

static void CreateGameClientSocket(a8::TcpSession **p)
{
    *p = new GCClientSession();
}

static void GSListeneron_error(a8::TcpListener*, int type, int errorid)
{
    f8::UdpLog::Instance()->Debug("GCListeneron_error %d %d", {type, errorid});
}

void GCListener::Init()
{
    tcp_listener_ = new a8::TcpListener();
    tcp_listener_->on_create_client_socket = CreateGameClientSocket;
    tcp_listener_->on_error = GSListeneron_error;

    tcp_listener_->bind_address = "0.0.0.0";
    tcp_listener_->bind_port = JsonDataMgr::Instance()->GetConf()->At("listen_port")->AsXValue();
    tcp_listener_->Open();
    f8::MsgQueue::Instance()->RegisterCallBack
        (
         IM_HandleRedirect,
         [this] (const a8::Args& args)
         {
             int socket_handle = args.Get<int>(0);
             std::string url = args.Get<std::string>(1);
             std::string query_str = args.Get<std::string>(2);
             websocket_url_hash_[socket_handle] = std::make_tuple(url, query_str);
         });
    f8::MsgQueue::Instance()->RegisterCallBack
        (
         IM_ClientSocketDisconnect,
         [this] (const a8::Args& args)
         {
             int socket_handle = args.Get<int>(0);
             websocket_url_hash_.erase(socket_handle);
         });
}

void GCListener::UnInit()
{
    delete tcp_listener_;
    tcp_listener_ = nullptr;
}

void GCListener::SendBuf(unsigned short sockhandle, char* buf, int buflen)
{
    tcp_listener_->SendClientMsg(sockhandle, buf, buflen);
}

void GCListener::SendText(unsigned short sockhandle, const std::string& text)
{
    tcp_listener_->SendClientMsg(sockhandle, text.data(), text.size());
}

void GCListener::ForceCloseClient(unsigned short sockhandle)
{
    tcp_listener_->ForceCloseClient(sockhandle);
}

void GCListener::MarkClient(unsigned short sockhandle, bool is_active)
{
    tcp_listener_->MarkClient(sockhandle, is_active);
}

long long GCListener::GetSendNodeNum()
{
    return tcp_listener_->send_node_num;
}

long long GCListener::GetSentBytesNum()
{
    return tcp_listener_->sent_bytes_num;
}

bool GCListener::GetWebSocketUrl(int socket_handle, std::string& url, std::string& query_str)
{
    auto itr = websocket_url_hash_.find(socket_handle);
    if (itr != websocket_url_hash_.end()) {
        url = std::get<0>(itr->second);
        query_str = std::get<1>(itr->second);
        return true;
    } else {
        return false;
    }
}

int GCListener::GetSocketCount()
{
    return tcp_listener_->GetClientSocketCount();
}
